import pathlib

import typer
from ambix.migration_home import MigrationHome


def render(
    alembic_home_dir: pathlib.Path = typer.Option(
        "etc/alembic/versions",
        exists=True,
        help="The directory where the alembic migrations live",
    ),
):
    home = MigrationHome(alembic_home_dir)
    graph = home.generate_dependency_graph()
    print("strict digraph {")
    for (key, value) in graph.items():
        for child in value:
            if child is None:
                continue
            print(f"  {key} -> {child}")
    print("}")



import collections
import itertools


def iter_not_string(val):
    """
    Returns true if the argument is iterable but not a string
    """
    return isinstance(val, collections.Iterable) and not isinstance(val, str)


def compose_iters(*args):
    """
    flattens iterables passed as arguments into a single iterable collection
    """
    return itertools.chain(*(i if iter_not_string(i) else [i] for i in args))


def pairwise(seq):
    return zip(seq, seq[1:])



import pathlib
import typing

import toposort
from ambix.exceptions import AmbixError
from ambix.migration_script import MigrationScript
from ambix.utilities import compose_iters, pairwise


class MigrationHome:
    def __init__(self, home_path: pathlib.Path):
        AmbixError.require_condition(
            home_path.exists(),
            f"migration home path '{home_path}' doesn't exist",
        )
        self.home_path = home_path

        self.migrations = {}
        for migration_file in self.home_path.iterdir():
            if not migration_file.suffix == ".py":
                continue
            script = MigrationScript.parse_file(migration_file)
            self.migrations[script.revision] = script

    def generate_dependency_graph(self):
        return {k: set(compose_iters(v.down_revision)) for (k, v) in self.migrations.items()}

    def flatten(self):
        dependencies = self.generate_dependency_graph()
        flattened = toposort.toposort_flatten(dependencies)
        pairs = pairwise(flattened)
        for (new_down, rev) in pairs:
            self.migrations[rev].change_down_revision(new_down)

    def prune(self, rev: str, leave_migration: bool = False):
        AmbixError.require_condition(
            rev in self.migrations,
            f"Couldn't find a revision for {rev}",
        )
        rev_downs = self.migrations[rev].get_down_revisions()
        for migration in self.migrations.values():
            mig_downs = migration.get_down_revisions()
            if len(mig_downs) == 0:
                continue
            elif rev in mig_downs:
                migration.change_down_revision(*(rev_downs | mig_downs - set([rev])))
        if not leave_migration:
            self.migrations.pop(rev).file_path.unlink()

    def heads(self):
        graph = self.generate_dependency_graph()
        possible_heads = set(graph.keys())
        for down_rev_set in graph.values():
            possible_heads = possible_heads - down_rev_set
        return possible_heads

    def move(self, rev: str, *new_bases: typing.List[str]):
        """
        Moves a single revision to descend from new_bases. Leaves any
        descendant migrations of rev in place by first pruning rev
        """
        if len(new_bases) == 0:
            new_bases = self.heads()
        self.prune(rev, leave_migration=True)
        self.migrations[rev].change_down_revision(*new_bases)

    def ancestors(self, rev: str) -> typing.Iterator[MigrationScript]:
        """
        provides a generator that finds all ancestors of a rev by performing a
        depth-first search
        """
        visited = set()
        yield from self._ancestors(rev, visited)

    def _ancestors(self, rev: str, visited: typing.List[MigrationScript]) -> typing.Iterator[MigrationScript]:
        child = self.migrations[rev]
        parents = child.get_down_revisions()
        for parent in sorted(parents):
            if parent in visited:
                continue
            visited.add(parent)
            yield from self._ancestors(parent, visited=visited)
            yield parent

    def rebase(self, rev: str, *new_bases: typing.List[str]):
        """
        Moves a revision and all of its descendents to be based on new_bases
        """
        AmbixError.require_condition(
            len(new_bases) > 0,
            "At least one new base must be supplied",
        )
        for base in new_bases:
            AmbixError.require_condition(
                rev not in self.ancestors(base),
                f"Cannot rebase. {rev} is an ancestor of new head {base}",
            )
        self.migrations[rev].change_down_revision(*new_bases)



import pathlib

import typer
from ambix.migration_home import MigrationHome


def flatten(
    alembic_home_dir: pathlib.Path = typer.Option(
        "etc/alembic/versions",
        exists=True,
        help="The directory where the alembic migrations live",
    ),
):
    home = MigrationHome(alembic_home_dir)
    home.flatten()



import typer
from ambix.commands import flatten, move, prune, rebase, render

cli = typer.Typer()
cli.command()(flatten.flatten)
cli.command()(move.move)
cli.command()(prune.prune)
cli.command()(rebase.rebase)
cli.command()(render.render)



import pathlib
import typing

import typer
from ambix.migration_home import MigrationHome


def move(
    revision: str = typer.Argument(
        ...,
        help="The revision hash to be moved",
    ),
    new_bases: typing.List[str] = typer.Argument(
        ...,
        help="The new base(s) to base the revision upon",
    ),
    alembic_home_dir: pathlib.Path = typer.Option(
        "etc/alembic/versions",
        exists=True,
        help="The directory where the alembic migrations live",
    ),
):
    home = MigrationHome(alembic_home_dir)
    home.move(revision, *new_bases)



from __future__ import annotations

import collections
import pathlib
import re
import typing

import redbaron
from ambix.exceptions import AmbixError


class MigrationScript:
    def __init__(self, revision: str = None, down_revision: str = None, file_path: pathlib.Path = None):
        self.revision = revision
        self.down_revision = down_revision
        self.file_path = file_path

    def get_down_revisions(self) -> typing.Set[str]:
        if self.down_revision is None:
            return set()
        elif isinstance(self.down_revision, collections.Iterable) and not isinstance(self.down_revision, str):
            return set(self.down_revision)
        else:
            return set([self.down_revision])

    @classmethod
    def parse_file(cls, file_path: pathlib.Path) -> MigrationScript:
        with AmbixError.handle_errors(f"failed to parse migration script {file_path}"):
            red = redbaron.RedBaron(file_path.read_text())
        self = cls(
            revision=red.find(
                "assignment",
                target=lambda t: t.value == "revision",
            ).value.to_python(),
            down_revision=red.find(
                "assignment",
                target=lambda t: t.value == "down_revision",
            ).value.to_python(),
            file_path=file_path,
        )
        return self

    def change_down_revision(self, *new_down_revisions: typing.List[str]):

        # TODO: don't change revision if it's the same

        if len(new_down_revisions) == 1:
            new_down_revisions = new_down_revisions[0]
        elif len(new_down_revisions) == 0:
            new_down_revisions = None

        with AmbixError.handle_errors("failed while changing down revision"):
            red = redbaron.RedBaron(self.file_path.read_text())
            red.find(
                "assignment",
                target=lambda t: t.value == "down_revision",
            ).value.replace(repr(new_down_revisions))

            docstring = red.find(
                "string",
                value=re.compile(
                    ".*Revises.*",
                    flags=re.MULTILINE | re.DOTALL,
                ),
            )
            if docstring is not None:
                if isinstance(new_down_revisions, collections.Iterable) and not isinstance(new_down_revisions, str):
                    rev_string = ", ".join(new_down_revisions)
                else:
                    rev_string = str(new_down_revisions)

                docstring.replace(
                    re.sub(
                        r"Revises:\s*\w+",
                        f"Revises: {rev_string}",
                        str(docstring),
                        flags=re.MULTILINE,
                    )
                )

            self.file_path.write_text(red.dumps())
            self.down_revision = new_down_revisions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ambix', 'ambix.commands']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.5.3,<0.6.0',
 'pendulum>=2.1.2,<3.0.0',
 'py-buzz>=2.1.3,<3.0.0',
 'redbaron>=0.9.2,<0.10.0',
 'toposort>=1.6,<2.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['ambix = ambix.cli:cli']}

setup_kwargs = {
    'name': 'ambix',
    'version': '1.0.0',
    'description': 'alembic history cleaning tool',
    'long_description': None,
    'author': 'Tucker Beck',
    'author_email': 'tucker.beck@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
